import logging
import pathlib
from abc import ABC, abstractmethod
from typing import Union

from .query import Query, QueryResult
from .stores import DataStore
from .stores import DataStoreManager

logger = logging.getLogger("gldb")


class GenericLinkedDatabase(ABC):

    @property
    @abstractmethod
    def store_manager(self) -> DataStoreManager:
        """Returns the store manager."""

    def __getitem__(self, store_name) -> DataStore:
        return self.store_manager[store_name]

    @abstractmethod
    def linked_upload(self, filename: Union[str, pathlib.Path]):
        """Uploads the file to both stores and links them."""

    def execute_query(self, store_name: str, query: Query) -> QueryResult:
        return self.store_manager.execute_query(store_name, query)



import logging
import pathlib
from logging.handlers import RotatingFileHandler

import appdirs

from ._version import __version__

USER_LOG_DIR = pathlib.Path(appdirs.user_log_dir(__package__, version=__version__))
USER_LOG_DIR.mkdir(parents=True, exist_ok=True)

DEFAULT_LOGGING_LEVEL = logging.DEBUG
_formatter = logging.Formatter(
    '%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d_%H:%M:%S')

_stream_handler = logging.StreamHandler()
_stream_handler.setLevel(DEFAULT_LOGGING_LEVEL)
_stream_handler.setFormatter(_formatter)

_file_handler = RotatingFileHandler(USER_LOG_DIR / f'{__package__}.log')
_file_handler.setLevel(logging.DEBUG)  # log everything to file!
_file_handler.setFormatter(_formatter)

logger = logging.getLogger(__package__)
logger.addHandler(_stream_handler)
logger.addHandler(_file_handler)

logger = logging.getLogger(__package__)
logger.addHandler(_stream_handler)
logger.addHandler(_file_handler)

from .gldb import GenericLinkedDatabase
from gldb.stores.rawdatastore import RawDataStore
from gldb.stores.metadatastore import RDFStore

__all__ = ['GenericLinkedDatabase', 'RDFStore', 'RawDataStore']



from abc import ABC, abstractmethod
from typing import Any


class AbstractQuery(ABC):

    @property
    @abstractmethod
    def description(self):
        """Returns a description of the query."""


class QueryResult:

    def __init__(self, query: AbstractQuery, result: Any):
        self.query = query
        self.result = result

    def __len__(self):
        return len(self.result)


class Query(AbstractQuery, ABC):

    @abstractmethod
    def execute(self, *args, **kwargs) -> QueryResult:
        """Executes the query."""



import json
import logging
import pathlib
import sys
import unittest
from typing import List, Union

import rdflib

from gldb import GenericLinkedDatabase
from gldb.query.rdfstorequery import SparqlQuery
from gldb.stores import DataStoreManager

logger = logging.getLogger("gldb")
logger.setLevel(logging.DEBUG)
for h in logger.handlers:
    h.setLevel(logging.DEBUG)

__this_dir__ = pathlib.Path(__file__).parent

sys.path.insert(0, str(__this_dir__))
from example_rdf_database import InMemoryRDFDatabase
from example_storage_db import CSVDatabase

from dataclasses import dataclass
from typing import Any, Dict


@dataclass(frozen=True)
class FederatedQueryResult:
    data: Any
    metadata: Dict


class GenericLinkedDatabaseImpl(GenericLinkedDatabase):

    def __init__(self):
        _store_manager = DataStoreManager()
        _store_manager.add_store("rdf_database", InMemoryRDFDatabase())
        _store_manager.add_store("csv_database", CSVDatabase())
        self._store_manager = _store_manager

    @property
    def store_manager(self) -> DataStoreManager:
        return self._store_manager

    # @property
    # def datastore(self) -> CSVDatabase:
    #     return self._storage_db

    def linked_upload(self, filename: Union[str, pathlib.Path]):
        raise NotImplemented("linked_upload not implemented")

    def get_temperature_data_by_date(self, date: str) -> List[FederatedQueryResult]:
        """High-level abstraction for user to find temperature data.
        It is a federated query that combines metadata and data from the RDF and CSV databases."""
        sparql_query = """
        PREFIX dcterms: <http://purl.org/dc/terms/>
        PREFIX dcat: <http://www.w3.org/ns/dcat#>

        SELECT ?dataset ?url
        WHERE {{
          ?dataset a dcat:Dataset .
          ?dataset dcterms:created "{date}" .
          ?dataset dcat:distribution ?distribution .
          ?distribution dcat:downloadURL ?url .
        }}
        """.format(date=date)
        results = self["rdf_database"].execute_query(SparqlQuery(sparql_query))

        result_data = [{str(k): parse_literal(v) for k, v in binding.items()} for binding in results.bindings]

        federated_query_results = []

        for res in result_data:
            filename = str(res["url"]).rsplit('/', 1)[-1]

            data = self["csv_database"].get_all(filename)

            # query all metadata for the dataset:
            metadata_sparql = """
            PREFIX dcterms: <http://purl.org/dc/terms/>
            PREFIX dcat: <http://www.w3.org/ns/dcat#>

            SELECT ?p ?o
            WHERE {{
              <{dataset}> ?p ?o .
            }}
            """.format(dataset=res["dataset"])
            metadata_result = self["rdf_database"].execute_query(SparqlQuery(metadata_sparql))
            dataset_result_data = [{str(k): v for k, v in binding.items()} for binding in
                                   metadata_result.bindings]
            metadata = {d["p"]: d["o"] for d in dataset_result_data}
            context = {"dcterms": "http://purl.org/dc/terms/", "dcat": "http://www.w3.org/ns/dcat#",
                       "ex": "https://example.org/"}

            g = rdflib.Graph()
            for k, v in metadata.items():
                g.add((rdflib.URIRef(res["dataset"]), rdflib.URIRef(k), v))
            jsonld = g.serialize(format="json-ld", context=context)
            jsonld_dict = json.loads(jsonld)
            for k, v in jsonld_dict.items():
                if isinstance(v, dict):
                    if len(v) == 1 and "@id" in v:
                        jsonld_dict[k] = v["@id"]
            federated_query_results.append(FederatedQueryResult(data=data, metadata=jsonld_dict))
            # better convert metadata to json-ld string

        return federated_query_results


def parse_literal(literal):
    if isinstance(literal, rdflib.Literal):
        return literal.value
    if isinstance(literal, rdflib.URIRef):
        return str(literal)
    return literal


class TestVersion(unittest.TestCase):

    def test_concrete_impl(self):
        db = GenericLinkedDatabaseImpl()

        rdf_database = db["rdf_database"]
        csv_database = db["csv_database"]

        rdf_database.upload_file(__this_dir__ / "data/data1.jsonld")
        query = SparqlQuery("SELECT * WHERE {?s ?p ?o}", description="Selects all triples")
        self.assertEqual(query.description, "Selects all triples")
        res = rdf_database.execute_query(query)

        self.assertEqual(8, len(res))
        self.assertEqual(len(res.result), len(res))

        rdf_database.upload_file(__this_dir__ / "data/metadata.jsonld")

        csv_database.upload_file(__this_dir__ / "data/random_data.csv")
        csv_database.upload_file(__this_dir__ / "data/random_data.csv")
        csv_database.upload_file(__this_dir__ / "data/temperature.csv")
        csv_database.upload_file(__this_dir__ / "data/users.csv")

        data = db.get_temperature_data_by_date(date="2024-01-01")
        self.assertIsInstance(data, list)
        self.assertIsInstance(data[0], FederatedQueryResult)



import unittest

from gldb.query import Query, QueryResult


class TestVersion(unittest.TestCase):

    def test_query(self):
        class SQLQuery(Query):
            def __init__(self, query: str, description: str):
                self.query = query
                self._description = description

            @property
            def description(self):
                return self._description

            def execute(self, *args, **kwargs) -> QueryResult:
                return QueryResult(self, "result")

        q = SQLQuery("SELECT * FROM Customers;", "Get all customers")
        assert q.query == "SELECT * FROM Customers;"
        assert q.description == "Get all customers"

        res = q.execute()
        assert isinstance(res, QueryResult)



import pathlib
from abc import ABC, abstractmethod
from typing import Union

from .store import DataStore


class RawDataStore(DataStore, ABC):
    """Data store interface (concrete implementations can be sql or non sql databases)."""

    @abstractmethod
    def upload_file(self, filename: Union[str, pathlib.Path]):
        """Insert data into the data store."""
        pass



import pathlib
from abc import ABC, abstractmethod
from typing import Union

import rdflib

from .store import DataStore


class RDFStore(DataStore, ABC):
    """Graph database interface."""

    namespaces = {
        "ex": "https://example.org/",
        "afn": "http://jena.apache.org/ARQ/function#",
        "agg": "http://jena.apache.org/ARQ/function/aggregate#",
        "apf": "http://jena.apache.org/ARQ/property#",
        "array": "http://www.w3.org/2005/xpath-functions/array",
        "dcat": "http://www.w3.org/ns/dcat#",
        "dcterms": "http://purl.org/dc/terms/",
        "fn": "http://www.w3.org/2005/xpath-functions",
        "foaf": "http://xmlns.com/foaf/0.1/",
        "geoext": "http://rdf.useekm.com/ext#",
        "geof": "http://www.opengis.net/def/function/geosparql/",
        "gn": "http://www.geonames.org/ontology#",
        "graphdb": "http://www.ontotext.com/config/graphdb#",
        "list": "http://jena.apache.org/ARQ/list#",
        "local": "https://doi.org/10.5281/zenodo.14175299/",
        "m4i": "http://w3id.org/nfdi4ing/metadata4ing#",
        "map": "http://www.w3.org/2005/xpath-functions/map",
        "math": "http://www.w3.org/2005/xpath-functions/math",
        "ofn": "http://www.ontotext.com/sparql/functions/",
        "omgeo": "http://www.ontotext.com/owlim/geo#",
        "owl": "http://www.w3.org/2002/07/owl#",
        "path": "http://www.ontotext.com/path#",
        "prov": "http://www.w3.org/ns/prov#",
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        "rep": "http://www.openrdf.org/config/repository#",
        "sail": "http://www.openrdf.org/config/sail#",
        "schema": "https://schema.org/",
        "spif": "http://spinrdf.org/spif#",
        "sr": "http://www.openrdf.org/config/repository/sail#",
        "ssno": "https://matthiasprobst.github.io/ssno#",
        "wgs": "http://www.w3.org/2003/01/geo/wgs84_pos#",
        "xsd": "http://www.w3.org/2001/XMLSchema#"
    }

    @property
    @abstractmethod
    def graph(self) -> rdflib.Graph:
        pass

    @abstractmethod
    def upload_file(self, filename: Union[str, pathlib.Path]) -> bool:
        """Insert data into the data store."""
        pass



from abc import ABC

import rdflib

from gldb.query.query import Query


class RDFStoreQuery(Query, ABC):
    """RDF Store Query interface."""


class SparqlQuery(RDFStoreQuery):

    def __init__(self, sparql_query: str, description: str="", *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._description = description
        self.sparql_query = sparql_query

    @property
    def description(self):
        return self._description

    def __repr__(self):
        return f"{self.__class__.__name__}({self.sparql_query!r})"

    def __str__(self):
        return self.sparql_query

    def execute(self, graph: rdflib.Graph) -> rdflib.query.Result:
        return graph.query(self.sparql_query, *self._args, **self._kwargs)



import pathlib

import rdflib

from gldb.query.rdfstorequery import SparqlQuery
from gldb.stores import RDFStore


class InMemoryRDFDatabase(RDFStore):

    def __init__(self):
        self._filenames = []
        self._graphs = {}
        self._expected_file_extensions = {".ttl", ".rdf", ".jsonld"}

    @property
    def expected_file_extensions(self):
        return self._expected_file_extensions

    def execute_query(self, query: SparqlQuery):
        return query.execute(self.graph)

    def upload_file(self, filename) -> bool:
        filename = pathlib.Path(filename)
        if not filename.exists():
            raise FileNotFoundError(f"File {filename} not found.")
        if filename.suffix not in self._expected_file_extensions:
            raise ValueError(f"File type {filename.suffix} not supported.")
        self._filenames.append(filename.resolve().absolute())
        return True

    @property
    def graph(self) -> rdflib.Graph:
        combined_graph = rdflib.Graph()
        for filename in self._filenames:
            g = self._graphs.get(filename, None)
            if not g:
                g = rdflib.Graph()
                g.parse(filename)
                for s, p, o in g:
                    if isinstance(s, rdflib.BNode):
                        new_s = rdflib.URIRef(f"https://example.org/{s}")
                    else:
                        new_s = s
                    if isinstance(o, rdflib.BNode):
                        new_o = rdflib.URIRef(f"https://example.org/{o}")
                    else:
                        new_o = o
                    g.remove((s, p, o))
                    g.add((new_s, p, new_o))
                self._graphs[filename] = g
            combined_graph += g
        return combined_graph



import pathlib
from abc import ABC, abstractmethod
from typing import Dict
from typing import Union

from gldb.query import Query, QueryResult


class DataStore(ABC):
    """Store interface."""

    @abstractmethod
    def execute_query(self, query: Query) -> QueryResult:
        """Executes the query on the store."""

    @abstractmethod
    def upload_file(self, filename: Union[str, pathlib.Path]) -> bool:
        """Uploads a file to the store."""


class DataStoreManager:
    """Store manager that manages the interaction between stores."""

    def __init__(self):
        self.stores: Dict[str, DataStore] = {}

    def add_store(self, store_name: str, store: DataStore):
        """Add a new store to the manager."""
        self.stores[store_name] = store

    def get_store(self, store_name: str) -> DataStore:
        """Retrieve a store from the manager."""
        return self.stores[store_name]

    def __getitem__(self, store_name: str) -> DataStore:
        """Retrieve a store from the manager."""
        return self.stores[store_name]

    def execute_query(self, store_name: str, query: Query) -> QueryResult:
        """Executes a query on a specific store."""
        store = self.get_store(store_name)
        if store:
            return store.execute_query(query)
        raise ValueError(f"Store {store_name} not found.")

    def upload_file(self, store_name: str, filename: str):
        """Uploads a file to a specific store."""
        store = self.get_store(store_name)
        if store:
            return store.upload_file(filename)
        raise ValueError(f"Store {store_name} not found.")
